functions {
  /* calculate the absolute value of a - b in log-space with log(a)
     and log(b) given. Does so by squaring and taking the root, i.e.

     la = log(a)
     lb = log(b)

     sqrt( (a - b)^2 ) = sqrt( a^2 - 2 * a * b + b^2 )

     <=> 0.5 * log_diff_exp(log_sum_exp(2*la, 2*lb), log(2) + la + lb)
  */
  real log_diff_exp_abs(real la, real lb) {
    return(0.5 * log_diff_exp(log_sum_exp(2*la, 2*lb), log(2) + la + lb));
  }

  vector pk_1cmt_oral_analytic(vector tad, real ldose, real lka, real lCL, real lV) {
    real lk;
    real k;
    real ka;
    real lkdsq;
    real lscale;
    vector[num_elements(tad)] pk;

    lk <- lCL - lV;
    lkdsq  <- 2*log_diff_exp_abs(lka, lk);

    k  <- exp(lk);
    ka <- exp(lka);

    // lump together the overall scaling factor
    lscale <- ldose - lV + lka - lkdsq;

    // calculates the Bateman function
    // (exp(-k * t) - exp(-ka * t))/(ka-k)
    // in log space which requires careful sign-stable
    // calculations
    for(i in 1:num_elements(tad)) {
      pk[i] <- lscale +
        log_diff_exp(log_sum_exp(lka - k  * tad[i], lk - ka * tad[i]),
                     log_sum_exp(lka - ka * tad[i], lk - k  * tad[i]) );
    }

    return(pk);
  }
}
data {
  int J;
  vector<lower=0>[J] dose;
  int N[J];
  vector<lower=0>[sum(N)] dv;
  vector<lower=0>[sum(N)] time;
}
transformed data {
  int Ndv;
  vector[sum(N)] ldv;
  vector[J] ldose;

  Ndv <- sum(N);

  ldv <- log(dv);
  ldose <- log(dose);
}
parameters {
  // log population parameters
  real lCL;
  real lV;
  // note: constraints can depend on other parameters!
  // enfore that absorption is faster than elimination
  real<lower=lCL - lV> lka;


  // subject specific random effects
  matrix[2,J] xi_eta;
  cholesky_factor_corr[2] LR_eta;   // prior correlations (LKJ prior on cholesky correlation factor)
  vector<lower=0>[2] sigma_eta;     // prior scales

  real<lower=0> sigma_y;
}
transformed parameters {
  matrix[2,J] Eta;
  vector[Ndv] mu;
  vector[2] theta;

  // combine parameters into vector for convenient manipulations
  theta[1] <- lCL;
  theta[2] <- lV;

  // for correlated random-effects
  Eta <- rep_matrix(theta, J) + diag_pre_multiply(sigma_eta, LR_eta ) * xi_eta;

  // for uncorrelated random effects
  //Eta <- rep_matrix(theta, J) + diag_pre_multiply(sigma_eta, xi_eta);

  // loop over each subject and evaluate PK model, use ragged array
  // structures
  {
    int DV_index;
    DV_index <- 1;
    for(j in 1:J) {
      mu[DV_index : (DV_index + N[j] - 1)] <- pk_1cmt_oral_analytic(segment(time, DV_index, N[j]), ldose[j], lka, Eta[1,j], Eta[2,j]);
      DV_index <- DV_index + N[j];
    }
  }
}
model {
  lka ~ normal(log(1),   log(2)/1.96);
  lCL ~ normal(log(0.1), log(10)/1.96);
  lV  ~ normal(log(10),  log(10)/1.96);

  LR_eta ~ lkj_corr_cholesky(4.0);
  sigma_eta ~ lognormal(log(0.3), log(4)/1.96);

  to_vector(xi_eta) ~ normal(0,1);

  sigma_y ~ normal(0, 2);

  ldv ~ normal(mu, sigma_y);
}
generated quantities {
  real ka;
  real CL;
  real V;
  vector[Ndv] log_lik;
  vector[Ndv] ires;
  real rho;

  {
    matrix[2,2] Rho;
    Rho <- multiply_lower_tri_self_transpose(LR_eta);
    rho <- Rho[1,2];
  }

  ka   <- exp(lka);
  CL   <- exp(lCL);
  V    <- exp(lV);

  for(o in 1:Ndv) {
    log_lik[o] <- normal_log(ldv[o], mu[o], sigma_y);
    ires[o] <- (mu[o] - ldv[o])/sigma_y ;
  }
}
