functions {

  // ODE RHS, dy_dt = f(t,y,theta)
  real[] pk_1cmt_oral_ode(real t, real[] y, real[] theta, real[] x_r, int[] x_i) {
    real dy_dt[2];
    real ka;
    real ke;

    ka <- theta[1];
    ke <- theta[2];

    dy_dt[1] <- - ka * y[1];
    dy_dt[2] <-   ka * y[1] - ke * y[2];

    return(dy_dt);
  }

  real[,] pk_1cmt_oral(vector tad, real dose, real lka, real lCL, real lV, real[] x_r) {
    // variables declared within a function are always treated as
    // "var" variables ...
    real V;
    real pk[num_elements(tad),2];
    real y0[2];
    real theta[2];
    // .. with the exception of integers which are never var
    int x_i[0];

    y0[1] <- dose;
    y0[2] <- 0;

    theta[1] <- exp(lka);
    theta[2] <- exp(lCL - lV);
    V <- exp(lV);

    // x_r is treated as data as it comes from external and was
    // declared in the transformed data block
    pk <- integrate_ode(pk_1cmt_oral_ode, y0, 0, to_array_1d(tad), theta, x_r, x_i);

    // note that y0 is treated as "var", but gradient information is
    // not needed since the initial condition is independent of any
    // parameters, could be optimized - how?

    // new in 2.10 is the ability to specify relative+absolute tolerance
    // and maximal number of steps
    //pk <- integrate_ode_rk45(pk_1cmt_oral_ode, y0, 0, to_array_1d(tad), theta, x_r, x_i, 1e-5, 1e-5, 1e3);

    // ODE returns amounts, but we have observed concentrations
    for (i in 1:num_elements(tad)) {
      pk[i,2] <- pk[i,2] / V;
    }

    return (pk);
  }
}
data {
  int J;
  vector<lower=0>[J] dose;
  int N[J];
  vector<lower=0>[sum(N)] dv;
  vector<lower=0>[sum(N)] time;
}
transformed data {
  int Ndv;
  vector[sum(N)] ldv;
  real x_r[0];

  Ndv <- sum(N);

  ldv <- log(dv);
}
parameters {
  // log population parameters
  real lCL;
  real lV;
  // note: constraints can depend on other parameters!
  // enfore that absorption is faster than elimination
 real<lower=lCL - lV> lka;

  // subject specific random effects
  matrix[2,J] xi_eta;
  cholesky_factor_corr[2] LR_eta;   // prior correlations (LKJ prior on cholesky correlation factor)
  vector<lower=0>[2] sigma_eta;     // prior scales

  real<lower=0> sigma_y;
}
transformed parameters {
  matrix[2,J] Eta;
  vector[Ndv] mu;
  vector[2] theta;

  // combine parameters into vector for convenient manipulations below
  theta[1] <- lCL;
  theta[2] <- lV;

  // for correlated random-effects
  Eta <- rep_matrix(theta, J) + diag_pre_multiply(sigma_eta, LR_eta ) * xi_eta;

  // for uncorrelated random effects
  //Eta <- rep_matrix(theta, J) + diag_pre_multiply(sigma_eta, xi_eta);

  // loop over each subject and evaluate PK model, use ragged array
  // structures
  {
    int DV_index;
    DV_index <- 1;
    for (j in 1:J) {
      real pk[N[j], 2];
      pk <- pk_1cmt_oral(segment(time, DV_index, N[j]), dose[j], lka, Eta[1,j], Eta[2,j], x_r);

      for (i in 1:N[j])
        mu[DV_index + i - 1] <- log(pk[i,2]);

      DV_index <- DV_index + N[j];
    }
  }
}
model {
  lka ~ normal(log(1),   log(2)/1.96);
  lCL ~ normal(log(0.1), log(10)/1.96);
  lV  ~ normal(log(10),  log(10)/1.96);

  LR_eta ~ lkj_corr_cholesky(4.0);
  sigma_eta ~ lognormal(log(0.3), log(4)/1.96);

  to_vector(xi_eta) ~ normal(0,1);

  sigma_y ~ normal(0, 2);

  ldv ~ normal(mu, sigma_y);
}
generated quantities {
  real ka;
  real CL;
  real V;
  vector[Ndv] log_lik;
  vector[Ndv] ires;
  real rho;

  {
    matrix[2,2] Rho;
    Rho <- multiply_lower_tri_self_transpose(LR_eta);
    rho <- Rho[1,2];
  }

  ka   <- exp(lka);
  CL   <- exp(lCL);
  V    <- exp(lV);

  for(o in 1:Ndv) {
    log_lik[o] <- normal_log(ldv[o], mu[o], sigma_y);
    ires[o] <- (mu[o] - ldv[o])/sigma_y ;
  }
}
